import("pathfinder.road", "RoadPathFinder", 4);  // binary_heap
import("pathfinder.road", "RoadPathFinder6", 6); // fibonacci_heap
import("pathfinder.road", "RoadPathFinder7", 7); // native_heap
import("pathfinder.road", "RoadPathFinder8", 8); // priority_queue

class MainClass extends AIController
{
	_current_tick = null;

	function Start()
	{
		AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);

		local town_list = AITownList();
		local selected_heap = AIController.GetSetting("select_heap");
		local build_signs = AIController.GetSetting("build_signs");
		local build_roads = AIController.GetSetting("build_roads");
		local offsets = [AIMap.GetTileIndex(0, 1), AIMap.GetTileIndex(0, -1), AIMap.GetTileIndex(1, 0), AIMap.GetTileIndex(-1, 0)];
		
		local test_label = "";
		switch (selected_heap) {
			case 0: {
				test_label = "pathfinder.road 4 / graph.aystar 6 / queue.binary_heap 1";
				break;
			}
			case 1: {
				test_label = "pathfinder.road 6 (4) / graph.aystar 9 (6) / queue.fibonacci_heap 3";
				break;
			}
			case 2: {
				test_label = "pathfinder.road 7 (4) / graph.aystar 10 (6) / queue.native_heap 1";
				break;
			}
			case 3: {
				test_label = "pathfinder.road 8 (4) / graph.aystar 11 (6) / queue.priority_queue 2";
				break;
			}
		}
		AILog.Info(test_label);

//		while (true)
		{
			local town2_list = AITownList();
			for (local town = town_list.Begin(); !town_list.IsEnd(); town = town_list.Next()) {
				local town1_location = AITown.GetLocation(town);
				for (local town2 = town2_list.Begin(); !town2_list.IsEnd(); town2 = town2_list.Next()) {
					if (town == town2) continue;
					local town2_location = AITown.GetLocation(town2);

					local pathfinder;
					switch (selected_heap) {
						case 0: {
							pathfinder = RoadPathFinder();
							break;
						}
						case 1: {
							pathfinder = RoadPathFinder6();
							break;
						}
						case 2: {
							pathfinder = RoadPathFinder7();
							break;
						}
						case 3: {
							pathfinder = RoadPathFinder8();
							break;
						}
					}
					
					switch (selected_heap) {
						case 0:
						case 1:
						case 2:
						case 3: {
							pathfinder.InitializePath([town1_location], [town2_location]);
							break;
						}
					}
					
					local cur_tick = AIController.GetTick();
					if (_current_tick == null) _current_tick = cur_tick;
					local path = false;
					switch (selected_heap) {
						case 0:
						case 1:
						case 2:
						case 3: {
							path = pathfinder.FindPath(-1);
							if (path == null) AILog.Info("path == null");
							break;
						}
					}
					AILog.Info(AITown.GetName(town) + " -> " + AITown.GetName(town2) + " := " + AIMap.DistanceManhattan(town1_location, town2_location) + " tiles // " + (AIController.GetTick() - cur_tick) + " ticks");
					if (!build_roads && !build_signs) continue;

					switch (selected_heap) {
						case 0:
						case 1:
						case 2:
						case 3: {
							local last_node = null;
							while (path != null) {
								local par = path.GetParent();
								if (par != null) {
									if (AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) == 1) {
										if (!AIRail.IsRailTile(par.GetTile()) || (AITestMode() && !AIRoad.BuildRoad(path.GetTile(), par.GetTile()) || AIError.GetLastErrorString() == "ERR_ALREADY_BUILT") && !AIRail.IsLevelCrossingTile(par.GetTile())) {
											local counter = 0;
											if (build_signs) AISign.BuildSign(path.GetTile(), selected_heap.tostring());
											if (build_roads) {
												if (!AIRoad.BuildRoad(path.GetTile(), par.GetTile())) {
													// todo
												}
											}
										} else {
											// todo
										}
									} else {
										if (!AIBridge.IsBridgeTile(path.GetTile()) || !AITunnel.IsTunnelTile(path.GetTile())) {
											if (AIRoad.IsRoadTile(path.GetTile()) && !AIRoad.IsDriveThroughRoadStationTile(path.GetTile()) && AITunnel.GetOtherTunnelEnd(path.GetTile()) == par.GetTile()) {
												if (build_roads) {
													if (!AITile.DemolishTile(path.GetTile())) {
														// todo
													}
												}
											}
										}
										if (AITunnel.GetOtherTunnelEnd(path.GetTile()) == par.GetTile()) {
											if (build_signs) AISign.BuildSign(path.GetTile(), selected_heap.tostring());
											if (build_roads) {
												if (!AITunnel.BuildTunnel(AIVehicle.VT_ROAD, path.GetTile())) {
													// todo
												}
											}
										} else {
											local bridge_length = AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) + 1;
											local bridge_list = AIBridgeList_Length(bridge_length);
											for (local bridge = bridge_list.Begin(); !bridge_list.IsEnd(); bridge = bridge_list.Next()) {
												bridge_list.SetValue(bridge, AIBridge.GetMaxSpeed(bridge));
											}
											bridge_list.Sort(AIList.SORT_BY_VALUE, AIList.SORT_DESCENDING);
											if (build_signs) AISign.BuildSign(path.GetTile(), selected_heap.tostring());
											if (build_roads) {
												if (!AIBridge.BuildBridge(AIVehicle.VT_ROAD, bridge_list.Begin(), path.GetTile(), par.GetTile())) {
													// todo
												}
											}
										}
									}
								}
								last_node = path.GetTile();
								path = par;
							}
							break;
						}
					}

				}
			}
		}
		local end_tick = AIController.GetTick();
		local ticks = end_tick - _current_tick;
		AILog.Warning("Finished in " + ticks + " ticks");
	}
}
